/************************************************************************/
/*                                                                      */
/*         XX      XX W           W  III  NN    N    GGGG               */
/*           XX  XX    W         W    I   N N   N   G                   */
/*             XX       W   W   W     I   N  N  N   G  GGG              */
/*           XX  XX      W W W W      I   N   N N   G    G              */
/*         XX      XX     W   W      III  N    NN    GGGG               */
/*                                                                      */
/*  QtAirfoil Version 0.1        U.Lehnert                   3 /2010    */
/*                                                                      */
/*  Berechnung der Skelettlinie von Tragflügel-Profilen,                */
/*  deren aerodynamischen Kenngrößen                                    */
/*  und deren Unterteilung in Segmente zur Benutzung mit XWing          */
/*                                                                      */
/************************************************************************/


/*

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include <QtGui>
#include <math.h>
#include "airfoil.h"
#include "mainwindow.h"
#include "diagramscene.h"
#include "diagramview.h"
#include "spline.h"

MainWindow::MainWindow()
{
    createActions();
    createMenus();
    createStatusBar();
    createGridGroupBox();
    setCentralWidget(gridGroupBox);
    setCurrentFile("");
    afValid = FALSE;
}

void MainWindow::createPanelingGroup()
{
  panelingGroup = new QGroupBox();
  QGridLayout *panelingLayout = new QGridLayout();
  QLabel *npanLabel = new QLabel(tr("number of panels"));
  panelingLayout->addWidget(npanLabel,1,1);
  QLabel *ledLabel = new QLabel(tr("leading edge density"));
  panelingLayout->addWidget(ledLabel,2,1);
  QLabel *tedLabel = new QLabel(tr("trailing edge density"));
  panelingLayout->addWidget(tedLabel,3,1);

  npanSpinBox = new QSpinBox;
  npanSpinBox->setRange(3, 999);
  npanSpinBox->setValue(8);
  npanSpinBox->setSingleStep(1);
  connect(npanSpinBox, SIGNAL(editingFinished()), this, SLOT(panelingChanged()));
  panelingLayout->addWidget(npanSpinBox,1,2);
  
  ledSpinBox = new QDoubleSpinBox;
  ledSpinBox->setRange(1.0, 10.0);
  ledSpinBox->setValue(3.0);
  ledSpinBox->setSingleStep(0.5);
  connect(ledSpinBox, SIGNAL(editingFinished()), this, SLOT(panelingChanged()));
  panelingLayout->addWidget(ledSpinBox,2,2);
  
  tedSpinBox = new QDoubleSpinBox;
  tedSpinBox->setRange(1.0, 10.0);
  tedSpinBox->setValue(2.0);
  tedSpinBox->setSingleStep(0.5);
  connect(tedSpinBox, SIGNAL(editingFinished()), this, SLOT(panelingChanged()));
  panelingLayout->addWidget(tedSpinBox,3,2);

  panelingDisplay = new (QTextEdit);
  parameterDisplay->setReadOnly(TRUE);
  parameterDisplay->setPlainText(tr(""));
  panelingLayout->addWidget(panelingDisplay,1,3,4,1);

  panelingLayout->setColumnMinimumWidth (1, 80);
  panelingLayout->setColumnMinimumWidth (2, 80);
  panelingLayout->setColumnMinimumWidth (3, 200);
  panelingLayout->setRowStretch(1, 1);
  panelingLayout->setRowStretch(2, 1);
  panelingLayout->setRowStretch(3, 1);
  panelingLayout->setRowStretch(4, 10);
  panelingLayout->setColumnStretch(1, 10);
  panelingLayout->setColumnStretch(2, 10);
  panelingLayout->setColumnStretch(3, 20);
  panelingGroup->setLayout(panelingLayout);
}

void MainWindow::createModifyGroup()
{
  modifyGroup = new QGroupBox();
  QGridLayout *modifyLayout = new QGridLayout();
  QLabel *alpha0Label = new QLabel(tr("alpha0 [deg]"));
  modifyLayout->addWidget(alpha0Label,1,1);
  QLabel *cm0Label = new QLabel(tr("cm0"));
  modifyLayout->addWidget(cm0Label,2,1);

  alpha0SpinBox = new QDoubleSpinBox;
  alpha0SpinBox->setRange(-10.0, 10.0);
  alpha0SpinBox->setDecimals(4);
  alpha0SpinBox->setSingleStep(0.1);
  modifyLayout->addWidget(alpha0SpinBox,1,2);
  
  cm0SpinBox = new QDoubleSpinBox;
  cm0SpinBox->setRange(-0.2, 0.2);
  cm0SpinBox->setDecimals(4);
  cm0SpinBox->setSingleStep(0.01);
  modifyLayout->addWidget(cm0SpinBox,2,2);
  
  modifyButton = new QPushButton(tr("Modify"));
  connect(modifyButton, SIGNAL(pressed()), this, SLOT(doModify()));
  modifyLayout->addWidget(modifyButton,3,2);

  modifyLayout->setColumnMinimumWidth (1, 100);
  modifyLayout->setColumnMinimumWidth (2, 100);
  modifyLayout->setRowStretch(1, 1);
  modifyLayout->setRowStretch(2, 1);
  modifyLayout->setRowStretch(3, 1);
  modifyLayout->setRowStretch(4, 10);
  modifyLayout->setColumnStretch(1, 10);
  modifyLayout->setColumnStretch(2, 10);
  modifyGroup->setLayout(modifyLayout);
}

void MainWindow::createGridGroupBox()
{
  gridGroupBox = new QGroupBox();
  layout = new QGridLayout;
  canvas = new DiagramScene;

  view = new DiagramView(canvas);
  layout->addWidget(view,1,1);

  parameterDisplay = new QTextEdit;
  parameterDisplay->setReadOnly(TRUE);
  parameterDisplay->setPlainText(tr(""));
  layout->addWidget(parameterDisplay, 1, 2);

  createPanelingGroup();
  layout->addWidget(panelingGroup, 2, 1);

  createModifyGroup();
  layout->addWidget(modifyGroup, 2, 2);

  camberlineDisplay = new QTextEdit;
  camberlineDisplay->setReadOnly(TRUE);
  camberlineDisplay->setPlainText(tr(""));
  layout->addWidget(camberlineDisplay, 3, 1, 1, 2);

  layout->setColumnMinimumWidth (1, 400);
  layout->setColumnMinimumWidth (2, 200);
  layout->setColumnStretch(1, 20);
  layout->setColumnStretch(2, 10);
  layout->setRowMinimumHeight (1, 150);
  layout->setRowMinimumHeight (2, 150);
  layout->setRowMinimumHeight (3, 60);
  layout->setRowStretch(1, 25);
  layout->setRowStretch(2, 15);
  layout->setRowStretch(3, 10);
  
  gridGroupBox->setLayout(layout);
}

void MainWindow::closeEvent(QCloseEvent *event)
{
    event->accept();
}

void MainWindow::open()
{
    QString fileName = QFileDialog::getOpenFileName(this,
        tr("Open airfoil datafile"), ".",
           tr("airfoil datafiles (*.dat)\n"
             "all files (*.*)"));
    if (!fileName.isEmpty())
      loadFile(fileName);
}

void MainWindow::about()
{
  QMessageBox::about(this, tr("About QtAirfoil"),
                     tr("<b>QtAirfoil</b> demonstrates how to "
               "write modern GUI applications using Qt."));
}

void MainWindow::panelingChanged()
{
  panelingDisplay->setPlainText(QString(""));
  if (afValid)
    {
      if (panCL!=0) delete panCL;
      int n = npanSpinBox->value();
      panCL = new Camberline(n, ledSpinBox->value(),
                                tedSpinBox->value());
      panCL->computePoints(curAirfoil);
      panelingRegenerateDisplay();
    }
  else
    {
      camberlineDisplay->setPlainText(QString("no data"));
    };
}

void MainWindow::panelingRegenerateDisplay()
{
  int n = npanSpinBox->value();
  panCL->computeCoeff(&cur_cl0, &cur_alfa0, &cur_cm0,
                       &cur_dclda, &cur_asf, &cur_clsf);
  panelingDisplay->append(QString::number(panCL->npoints()) + QString(" points"));
  panelingDisplay->append(QString("cl0 = ") + QString::number(cur_cl0,'f',3));
  panelingDisplay->append(QString("alpha0 = ") +
      QString::number(180.0/M_PI*cur_alfa0,'f',3) + QString(" deg"));
  panelingDisplay->append(QString("cm0 = ") + QString::number(cur_cm0,'f',4));
  panelingDisplay->append(QString("d(cl)/da = ") + QString::number(cur_dclda,'f',3));
  panelingDisplay->append(QString("alpha(sf) = ") +
      QString::number(180.0/M_PI*cur_asf,'f',3) + QString(" deg"));
  panelingDisplay->append(QString("cl(sf) = ") + QString::number(cur_clsf,'f',3));
      
  QString *dxs = new QString(" ");
  QString *als = new QString(" ");
  double alpha, dx;
  for (int i=0; i<n; i++) {
    dx = panCL->xpoint(i+1)-panCL->xpoint(i);
    alpha = 180.0/M_PI*atan((panCL->ypoint(i)-panCL->ypoint(i+1))/dx);
    dxs->append(QString(" %1").arg(dx,9,'f',4));
    als->append(QString(" %1").arg(alpha,9,'f',4));
  };
  QFont *font = new QFont("Courier",9);
  camberlineDisplay->setCurrentFont(*font);
  camberlineDisplay->setPlainText(*dxs);
  camberlineDisplay->append(*als);
      
  canvas->clear();
  canvas->plotAirfoil(curAirfoil);
  canvas->plotCamberline(curCL);
  canvas->plotPaneling(panCL);
}

void MainWindow::doModify()
{
  int i, n;
  double x, y;
  double al0a, al0b;
  double cm0a, cm0b;
  double d1, d2, d3, d4;
  double al0ist, cm0ist;
  double al0soll, cm0soll;
  double a, b;
  if (afValid) {
    // create two new camber lines to be added to the existing one
    n = npanSpinBox->value();
    Camberline *cla = new Camberline(n, ledSpinBox->value(),
                           tedSpinBox->value());
    Camberline *clb = new Camberline(n, ledSpinBox->value(),
                           tedSpinBox->value());
    cla->sety(0,0.0);
    clb->sety(0,0.0);
    for (i=1; i<n; i++)
      {
        x = cla->xpoint(i);
        cla->sety(i,0.04*x*(1.0-x));          // parabolic camber
        clb->sety(i,0.1*x*(1.0-x)*(0.5-x));   // cubic S-shape
      };
    cla->sety(n,0.0);
    clb->sety(n,0.0);
    panCL->computeCoeff(&d1, &al0ist, &cm0ist, &d2, &d3, &d4);
    panelingDisplay->setPlainText(QString("modified paneling"));
    panelingDisplay->append(QString("start  :  a0=%1  cm0=%2").
        arg(180.0/M_PI*al0ist,8,'f',3).arg(cm0ist,8,'f',4));
    cla->computeCoeff(&d1, &al0a, &cm0a, &d2, &d3, &d4);
    clb->computeCoeff(&d1, &al0b, &cm0b, &d2, &d3, &d4);
    al0soll = M_PI/180.0*alpha0SpinBox->value();
    cm0soll = cm0SpinBox->value();
    panelingDisplay->append(QString("target :  a0=%1  cm0=%2").
        arg(180.0/M_PI*al0soll,8,'f',3).arg(cm0soll,8,'f',4));
    // solve linear equation system
    // for coefficients a and b
    // al0(soll) = al0(ist) + a * al0(a) + b * al0(b)
    // cm0(soll) = cm0(ist) + a * cm0(a) + b * cm0(b)
    a = (al0b*(cm0ist-cm0soll)-cm0b*(al0ist-al0soll))/
        (al0a*cm0b-al0b*cm0a);
    b = (al0a*(cm0ist-cm0soll)-cm0a*(al0ist-al0soll))/
        (al0b*cm0a-al0a*cm0b);
    panelingDisplay->append(QString("coefficients :  a=%1  b=%2").
        arg(a,8,'f',4).arg(b,8,'f',4));
    for (i=1; i<n; i++)
      {
        y = panCL->ypoint(i);
        y += a*cla->ypoint(i);
        y += b*clb->ypoint(i);
        panCL->sety(i,y);
      };
      panelingRegenerateDisplay();
    delete cla;
    delete clb;
  };
}

void MainWindow::createActions()
{
    openAct = new QAction(tr("&Open..."), this);
    openAct->setShortcut(tr("Ctrl+O"));
    openAct->setStatusTip(tr("Open an airfoil *.dat file"));
    connect(openAct, SIGNAL(triggered()), this, SLOT(open()));

    exitAct = new QAction(tr("E&xit"), this);
    exitAct->setShortcut(tr("Ctrl+Q"));
    exitAct->setStatusTip(tr("Exit QtAirfoil"));
    connect(exitAct, SIGNAL(triggered()), this, SLOT(close()));

    aboutAct = new QAction(tr("&About"), this);
    aboutAct->setStatusTip(tr("Show QtAirfoil's About box"));
    connect(aboutAct, SIGNAL(triggered()), this, SLOT(about()));

    aboutQtAct = new QAction(tr("About &Qt"), this);
    aboutQtAct->setStatusTip(tr("Show the Qt library's About box"));
    connect(aboutQtAct, SIGNAL(triggered()), qApp, SLOT(aboutQt()));
}

void MainWindow::createMenus()
{
    fileMenu = menuBar()->addMenu(tr("&File"));
    fileMenu->addAction(openAct);
    fileMenu->addAction(exitAct);

    menuBar()->addSeparator();

    helpMenu = menuBar()->addMenu(tr("&Help"));
    helpMenu->addAction(aboutAct);
    helpMenu->addAction(aboutQtAct);
}

void MainWindow::createStatusBar()
{
    statusBar()->showMessage(tr("Ready"));
}

void MainWindow::loadFile(const QString &fileName)
{
    double x, y;
    double xp[999];
    double yp[999];
    int n;

    QFile file(fileName);
    if (!file.open(QFile::ReadOnly | QFile::Text)) {
      QMessageBox::warning(this, tr("QtAirfoil"),
                             tr("Cannot read file %1:\n%2.")
                             .arg(fileName)
                             .arg(file.errorString()));
        return;
    }

    QTextStream in(&file);
    QApplication::setOverrideCursor(Qt::WaitCursor);

    QString descr(in.readLine());
    parameterDisplay->setPlainText(descr);

    n=0;
    do {
      in >> x;
      in >> y;
      xp[n]=x;
      yp[n]=y;
      in.readLine();
      n++;
    } while (!in.atEnd());
    parameterDisplay->append(QString::number(n) + QString(" points\n"));
    
    curAirfoil = new Airfoil(n, xp, yp);
    curAirfoil->normalize();
    afValid = TRUE;

    curCL = new Camberline(200,3.0,2.0);
    curCL->computePoints(curAirfoil);
    curCL->computeCoeff(&cur_cl0, &cur_alfa0, &cur_cm0,
                        &cur_dclda, &cur_asf, &cur_clsf);
    parameterDisplay->append(QString("cl0 = ") + QString::number(cur_cl0,'f',3));
    parameterDisplay->append(QString("alpha0 = ") +
        QString::number(180.0/M_PI*cur_alfa0,'f',3) + QString(" deg"));
    parameterDisplay->append(QString("cm0 = ") + QString::number(cur_cm0,'f',4));
    parameterDisplay->append(QString("d(cl)/da = ") + QString::number(cur_dclda,'f',3));
    parameterDisplay->append(QString("alpha(sf) = ") +
        QString::number(180.0/M_PI*cur_asf,'f',3) + QString(" deg"));
    parameterDisplay->append(QString("cl(sf) = ") + QString::number(cur_clsf,'f',3));

    alpha0SpinBox->setValue(180.0/M_PI*cur_alfa0);
    cm0SpinBox->setValue(cur_cm0);
    panelingChanged();

    QApplication::restoreOverrideCursor();

    setCurrentFile(fileName);
    statusBar()->showMessage(tr("File loaded"), 2000);
}

void MainWindow::setCurrentFile(const QString &fileName)
{
    curFile = fileName;
    parameterDisplay->document()->setModified(false);
    setWindowModified(false);

    QString shownName;
    if (curFile.isEmpty())
        shownName = "none";
    else
        shownName = strippedName(curFile);

    setWindowTitle(tr("%1 - %2").arg(tr("QtAirfoil")).arg(shownName));
}

QString MainWindow::strippedName(const QString &fullFileName)
{
    return QFileInfo(fullFileName).fileName();
}
